(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Formulas
open Types
open Declarations

exception Exc_assparse of string

(* fix in all quantifiers the kind of a declaration to ProcessKind *)
(* in case the type is Typeconst(r) where r is a rolename defined  *)
(* in the association list rolelist.                               *)
let fix_dl rolelist = 
  List.map (function (vname,(kind,typeexp)) as orig ->
	      match typeexp with
		  Typeconst(tname) ->
		    if List.mem_assoc tname rolelist
		    then (vname,(ProcessKind,typeexp))
		    else orig
		| _ -> orig)
let rec fix_process_kind rolelist = function
    Ftrue | Fatom _ | Fsecret _ | Fprocessat _ | Fscenarioat _ as f -> f
  | Fand(left,right) ->
      Fand(fix_process_kind rolelist left, fix_process_kind rolelist right)
  | For(left,right) ->
      For(fix_process_kind rolelist left, fix_process_kind rolelist right)
  | Fimplies(left,right) ->
      Fimplies(fix_process_kind rolelist left, fix_process_kind rolelist right)
  | Fnot form -> Fnot(fix_process_kind rolelist form)
  | Fbox form -> Fbox(fix_process_kind rolelist form)
  | Funique(decl,form) ->
      Funique(fix_dl rolelist decl, fix_process_kind rolelist form)
  | Fexists(decl,form) ->
      Fexists(fix_dl rolelist decl, fix_process_kind rolelist form)
  | Fforall(decl,form) ->
      Fforall(fix_dl rolelist decl, fix_process_kind rolelist form)

let parse_assertion infilename =
  let inbuf = Lexing.from_channel (open_in infilename)
  in 
    Lexer.register_filename inbuf infilename;
    try
      let (specname,publics,raw_initial,raw_always) as raw_assertion =
	Parser.assertion Lexer.next_token inbuf
      in
	Assertions.check_assertion_fragment raw_assertion;
	let (signature, axioms, rolelist, vars, scenario) =
	  try Specparser.parse_spec specname
	  with Specparser.Exc_specparser s ->
	    raise (Exc_assparse ("Error loading specification " ^ specname ^
				   " while checking assertion: " ^ s))
	in
	let complete_signature =
	  let (pcsig,hcsig)=
	    Signatures.construct_signature signature
	  in pcsig@hcsig
	and initial_env =
	  [("x_M", (Declarations.Logical, Types.Typeconst "message"))]
	and roles_toplevel_decls = Roles.gather_toplevel_declarations rolelist
	and roles_checkpoints = Roles.gather_checkpoints rolelist
	and scene_checkpoints = Scenarios.gather_checkpoints scenario
	in
	let initial = fix_process_kind roles_checkpoints raw_initial
	and always = fix_process_kind roles_checkpoints raw_always
	in
	  (try
	     Utils.checksubset publics (List.map fst complete_signature)
	   with
	       Utils.Missing e ->
		 raise (Exc_assparse
			  ("The function symbol " ^ e ^
			     " declared as public in the "
			   ^ "assertion but is not declared in "
			   ^ "the specification.")));
	  let initial_typed =
	    try
	      Formulas.type_formula
		complete_signature initial_env roles_toplevel_decls
		roles_checkpoints scene_checkpoints initial
	    with
		Formulas.Exc_formula s ->
		  raise (Exc_assparse
			   ("Type error in initial formula : " ^ s))
	  and always_typed =
	    try
	      Formulas.type_formula
		complete_signature initial_env roles_toplevel_decls 
		roles_checkpoints scene_checkpoints always
	    with
		Formulas.Exc_formula s ->
		  raise (Exc_assparse
			   ("Type error in safety formula : " ^ s))
	  in
	    (specname,publics,initial_typed,always_typed)
    with
	Lexer.Lex_error(s,p) ->
	  raise (Exc_assparse
		   ("Lexer error on character ``" ^ s ^ "'' at " ^ 
		      (Lexer.string_of_lexposition p )))
      | Parsing.Parse_error ->
	  raise (Exc_assparse
		   ("Parsing error before " ^
		      (Lexer.string_of_lexposition
			 (Lexing.lexeme_start_p inbuf))))
      | Assertions.Exc_assertion s ->
	  raise (Exc_assparse ("Assertion incorrect because: " ^ s))
	    
