(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Types
open Declarations

(* static analysis information *)

type var_kind =
  | Var_unknown    (* kind is not yet known *)
  | Var_value      (* can only be an l value *)
  | Var_logical    (* expression which is a logical variable *)
  | Var_mutable    (* expression which is a mutable variable *)
  | Var_of_process (* access to a variable of a process *)
  | Var_pid        (* variable denoting a process *)

type info_exp = {
  exp_type : type_expression;
  exp_varkind : var_kind;
  exp_hasvalue : bool
}

let undef_info_exp = {
  exp_type = Typeundef;
  exp_varkind = Var_unknown;
  exp_hasvalue = false
}

let declarationkind_to_varkind = function
  | Logical -> Var_logical
  | Mutable -> Var_mutable
  | ProcessKind -> Var_pid
   
(* universal type of expressions *)
type expression =
    Ident of info_exp * string
  | Int of info_exp * int
  | Apply of info_exp * string * expression list
  | Infix of info_exp * string * expression * expression
  | Inverse of info_exp * expression
  | Tuple of info_exp * expression list
  | Liste of info_exp * expression list
  | AssocList of info_exp * (expression*expression) list

let get_info_exp = function
    Ident(i,_) | Int(i,_) | Apply(i,_,_) | Infix(i,_,_,_) | Inverse(i,_)
  | Tuple(i,_) | Liste(i,_) | AssocList(i,_) -> i

type pattern = expression list


(************************** Conversion to strings ***********************)

let st showtypeflag info =
  if showtypeflag
  then "["^(string_of_type_expression info.exp_type)^"]"
  else ""

let rec string_of_expression showtype = function
  | Ident (info,s) -> (st showtype info)^s
  | Int (info,i) -> (st showtype info)^(string_of_int i)
  | Apply (info,f,l) -> (st showtype info)^f ^  
      (match l with
	   [] -> ""
	 | _  -> "(" ^ (string_of_expression_list showtype l) ^ ")")
  | Infix (info,op,e1,e2) ->
      "(" ^ (string_of_expression showtype e1) ^ " " ^
	(st showtype info)^op ^
      " " ^ (string_of_expression showtype e2) ^ ")"
  | Inverse (info,e) -> (st showtype info) ^ 
      "inv(" ^ (string_of_expression showtype e) ^ ")"
  | Tuple (info,l) -> (st showtype info) ^
"[" ^ (string_of_expression_list showtype l) ^ "]"
  | Liste (info,l) -> (st showtype info) ^
      "[[" ^ (string_of_expression_list showtype l) ^ "]]"
  | AssocList (info,l) -> (st showtype info) ^
      "{" ^ (string_of_expression_assoclist showtype l) ^ "}"
and string_of_expression_list showtype = function
    [] -> ""
  | h::r ->
      List.fold_left
	(fun s e -> s^", "^(string_of_expression showtype e) )
	(string_of_expression showtype h)
	r
and string_of_expression_assoclist showtype = function
  | [] -> ""
  | (h1,h2)::r ->
      List.fold_left
	(fun s (e1,e2) ->
	   s^", "^(string_of_expression showtype e1)^": "^
	     (string_of_expression showtype e2) )
	((string_of_expression showtype h1)^": "^
	   (string_of_expression showtype h2))
	r

let string_of_pattern showtype pat =
  Utils.string_of_string_list "; "
    (List.map (string_of_expression showtype) pat) 

(************ renaming of variables *************************************)

let rec rename_expr alpha = function
  | Ident (info,x) -> 
      (try
	 Ident (info,(List.assoc x alpha))
       with
	   Not_found -> Ident (info,x))
  | Int (info,i) -> Int (info,i)
  | Infix (info,op,e1,e2) ->
      Infix(info,op, rename_expr alpha e1, rename_expr alpha e2)
  | Inverse (info,e) -> Inverse (info,(rename_expr alpha e))
  | Tuple (info,l) -> Tuple (info,(List.map (rename_expr alpha) l)) 
  | Liste (info,l) -> Liste (info,(List.map (rename_expr alpha) l))
  | Apply (info,f,l) -> Apply(info,f,List.map (rename_expr alpha) l) 
  | AssocList (info,l) ->
      AssocList (info,(List.map
			 (function (l,r) ->
			    (rename_expr alpha l, rename_expr alpha r)))
						 l)

let rename_pattern ren = List.map (rename_expr ren)

(************   restricted forms of expressions   ***********************)

let is_l_expression = function
  |Ident _ | Apply _ -> true
  | _ -> false

(****************  collect integer constants  *****************************)

let int_constants e =
  let rec ic = function
    | Ident _ -> []
    | Int (_,i) -> [i]
    | Inverse (_,e) -> ic e
    | AssocList (_,l) ->
	(List.flatten (List.map ic (List.map fst l))) @
	  (List.flatten (List.map ic (List.map snd l)))
    | Apply (_,_,el) | Tuple (_,el) | Liste (_,el) ->
	List.flatten (List.map ic el)
    | Infix(_,_,e1,e2) -> (ic e1) @ (ic e2)
  in
    Utils.singlyfy (ic e)

(************** collect variables ***************************************)

let vars e =
  let rec vars_aux = function
    | Ident (info,name) -> 
	(match info.exp_varkind with
	   | Var_logical | Var_mutable | Var_of_process | Var_pid -> [name]
	   | Var_value -> []
	   | Var_unknown ->
	       failwith "vars_aux applied on identifier of unknown kind")
    | Int _ -> []
    | Inverse (_,e) -> vars_aux e
    | AssocList (_,l) ->
	(List.flatten (List.map vars_aux (List.map fst l))) @
	  (List.flatten (List.map vars_aux (List.map snd l)))
    | Apply (_,_,el) | Tuple (_,el) | Liste (_,el) ->
	List.flatten (List.map vars_aux el)
    | Infix(_,_,e1,e2) -> (vars_aux e1) @ (vars_aux e2)
	
  in
    Utils.singlyfy (vars_aux e)

let is_ground e = (List.length (vars e) = 0)

let mutable_vars e =
  let rec vars_aux = function
    | Ident (info,name) -> 
	(match info.exp_varkind with
	   | Var_mutable -> [name]
	   | Var_logical | Var_value | Var_of_process | Var_pid -> []
	   | Var_unknown ->
	       failwith "vars_aux applied on identifier of unknown kind")
    | Int _ -> []
    | Inverse (_,e) -> vars_aux e
    | AssocList (_,l) ->
	(List.flatten (List.map vars_aux (List.map fst l))) @
	  (List.flatten (List.map vars_aux (List.map snd l)))
    | Apply (_,_,el) | Tuple (_,el) | Liste (_,el) ->
	List.flatten (List.map vars_aux el)
    | Infix(_,_,e1,e2) -> (vars_aux e1) @ (vars_aux e2)
	
  in
    Utils.singlyfy (vars_aux e)


(*************************** Typing *************************************)
exception Exc_typeexpression of string

let pair_map f (l,r) = (f l), (f r)

let extract_type_expr = function
  | Ident(info,_) -> info.exp_type
  | Int(info,_) -> info.exp_type
  | Apply(info,_,_) -> info.exp_type
  | Infix(info,_,_,_) -> info.exp_type
  | Inverse(info,_) -> info.exp_type
  | Tuple(info,_) -> info.exp_type
  | Liste(info,_) -> info.exp_type
  | AssocList(info,_) -> info.exp_type


let rec type_expr sign env skenv r_env = function

  | Int (info,i) ->
      Int ({info with
	      exp_type = Typeconst "int";
	      exp_varkind = Var_value},
	   i)

  | Tuple (info,l) -> 
      let l_typed = List.map (type_expr sign env skenv r_env) l
      in Tuple(
	  {info with
	     exp_type = Typetuple (List.map extract_type_expr l_typed);
	     exp_varkind = Var_value},
	  l_typed)

  | Liste (info,l) -> 
      let l_typed = List.map (type_expr sign env skenv r_env) l
      in Liste(
	  {info with
	     exp_type = Typeapply("list",
				  [List.fold_left sup_type
				     Typeundef
				     (List.map extract_type_expr l_typed)]);
	     exp_varkind = Var_value},
	  l_typed)
	   
  | AssocList (info,l)  ->
      let l_typed =
	List.map (pair_map (type_expr sign env skenv r_env)) l
      in AssocList(
	  {info with
	     exp_type = Typeapply("table",
				  [List.fold_left sup_type
				     Typeundef
				     (List.map
					(function (x,_) -> extract_type_expr x)
					l_typed);
				   List.fold_left sup_type
				     Typeundef
				     (List.map
					(function (_,x) -> extract_type_expr x)
					l_typed)]);
	     exp_varkind = Var_value},
	  l_typed)

  | Ident (info,s) ->
      (try
	 let (s_kind,s_type) = List.assoc s env
	 in
	   Ident ({info with
		     exp_varkind = declarationkind_to_varkind s_kind;
		     exp_type = s_type},
		  s)
       with
	  Not_found -> 
	    try
	      let s_type = List.assoc s sign
	      in if arrow_free s_type
		then
		  Ident(
		    {info with
		       exp_varkind = Var_value;
		       exp_type = s_type},
		    s)
		else raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ s ^ "\" is used without arguments "
		      ^ "but is declared as of functional type "
		      ^ (string_of_type_expression s_type)))
	    with
		Not_found -> raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ s ^ "\" is not declared, " ^
		      "neither as a variable nor as a constant operation")))

  | Inverse (info,e) ->
      let e_typed = type_expr sign env skenv r_env e
      in let e_type = extract_type_expr e_typed
      in let e_inv_type  = 
	  (match e_type  with
	     | Typeconst "pubkey" -> Typeconst "privkey"
	     | Typeconst "privkey" -> Typeconst "pubkey"
	     | Typeconst "message" -> Typeconst "message"
	     | othertype ->
		 raise (Exc_typeexpression
			  ("Argument of Inverse operation of type "
			   ^ (string_of_type_expression othertype))))
      in Inverse(
	  {info with
	     exp_type = e_inv_type;
	     exp_varkind = Var_value},
	  e_typed)
      
  | Apply(info,f,l) ->
      (* f may be a skolem variable or a function *)
      let l_typed = List.map (type_expr sign env skenv r_env) l
      in let l_types = List.map extract_type_expr l_typed
      in
	(try 
	   let (from_type_list,kind,result_type) = List.assoc f skenv 
	     (* case: f is a skolem variable *)
	   in
	     (* the type of a skolem variable is always ground *)
	     if smaller_type (Typetuple l_types) (Typetuple from_type_list)
	     then
	       Apply(
		 {info with
		    exp_type = result_type;
		    exp_varkind = declarationkind_to_varkind kind },
		 f, l_typed)
	     else  raise
	       (Exc_typeexpression
		  ("the identifier \"" ^ f ^ "\" is declared as a " ^
		     "global variable with argument type " ^
		     (string_of_type_expression_list from_type_list) ^
		     ", which does not match the type " ^
		     (string_of_type_expression_list l_types)))
	 with Not_found ->
	   (* f is not a skolem variable, hence it must be a function  *)
	   (* from the signature.                                      *)
	   (try
	      match List.assoc f sign with
		  Typearrow(from_type, result_type) ->
		    let arguments_type =
		      (* the Typetuple constructor for argument types  *)
		      (* is only used for >= 2 arguments. If there is  *)
		      (* only one argument then we have to extract its *)
		      (* type, otherwise construct a typetuple.        *)
		      if List.length l_types = 1
		      then List.hd l_types
		      else Typetuple l_types
		    in
		      (* the type of a function may be polymorphic. We *)
		      (* hence have to match the argument type of f    *)
		      (* against the types of the actual arguments and *)
		      (* then to apply the matching subsitution on the *)
		      (* result type of f.                             *)
		      (try 
			 let result_type_inst =
			   apply_type_subst
			     (type_match from_type arguments_type) result_type
			 in
			   Apply(
			     {info with
				exp_type = result_type_inst;
				exp_varkind = Var_value},
			     f,l_typed)
		       with
			   Nomatch_type -> raise
			     (Exc_typeexpression
				("the identifier \"" ^ f ^
				   "\" is declared as a " ^
				   "function with argument type " ^
				   (string_of_type_expression from_type) ^
				   ", which does not match the type " ^
				   (string_of_type_expression arguments_type))))
		|  ftype -> raise
		     (Exc_typeexpression
			("the identifier \"" ^ f ^
			   "\"is used like a function " ^
			   "but has non-functional type " ^
			   (string_of_type_expression ftype)))
	    with
		Not_found -> raise
		  (Exc_typeexpression
		     ("the identifier \"" ^ f ^ "\" is used like a function "^
			"but is not declared"))))
	    
  | Infix(info,op,e1,e2) ->
      (* special case "." : this may be either a lookup in an     *)
      (* association list, or access to some toplevel variable    *)
      (* of a process. In order to find out we have to check      *)
      (* whether e1 is a process variable.                        *)

      (* e1 is always typed in the current environment, but we can  *)
      (* know how to type e2 only when we have determinend in which *)
      (* case we are.                                               *)

      let e1_typed = type_expr sign env skenv r_env e1
      in let e1_type = extract_type_expr e1_typed
      in
	if op = "." && (match e1_type
			with
			  | Typeconst rolename ->
			      List.mem_assoc rolename r_env
			  | _ -> false)
	then
	  (* We have an access to a toplevel identifier of a process *)
	  (* Then the resulting type is the type of e2 as given by  *)
	  (* the top-level declaration of the process e1.            *)
	  
	  match e2 with
	      Ident (e2_info,vid) -> 
		(match e1_type with
		     Typeconst rolename ->
		       (try 
			  let e2_type =
			    snd (List.assoc vid (List.assoc rolename r_env))
			  in let e2_typed =
			      Ident(
				{e2_info with
				   exp_type = e2_type;
				   exp_varkind = Var_of_process},
				vid)
			  in
			    Infix(
			      {info with
				 exp_type = e2_type;
				 exp_varkind = Var_of_process },
			      op, e1_typed, e2_typed)
			with
			    Not_found ->
			      raise
				(Exc_typeexpression
				   ("the identifier \""^vid^"\" is not a " ^
				      "toplevel identifier of role \""^
				      rolename^
				      "\"")))
		   | _ -> failwith "this cannot happen")
	    | _ -> raise
		(Exc_typeexpression
		   ((string_of_expression false e2) ^ "is not an identifier"))
	else
	  (* We have a lookup in an association list. In this case, *)
	  (* e2 must be typable in the current environment.         *)
	  let e2_typed = type_expr sign env skenv r_env e2
	  in let e2_type = extract_type_expr e2_typed
	  in
	    try
	      match List.assoc op sign with
		  Typearrow(from_type,result_type) ->
		    (* the type of an operator may be polymorphic. We *)
		    (* hence have to match the argument type of f     *)
		    (* against the types of the actual arguments and  *)
		    (* then to apply the matching substitution on the *)
		    (* result type of f.                              *)
		    (try 
		       let result_type_inst =
			 apply_type_subst
			   (type_match
			      from_type
			      (Typetuple [e1_type; e2_type]))
			   result_type
		       in
			 Infix(
			   {info with
			      exp_type = result_type_inst;
			      exp_varkind = Var_value},
			   op, e1_typed, e2_typed)
		     with
			 Nomatch_type -> raise
			   (Exc_typeexpression
			      ("the identifier \"" ^ op ^
				 "\" is declared as an " ^
				 "infix operator with argument argument type "
			       ^ (string_of_type_expression from_type) ^
				 " but is applied to arguments of " ^
				 "respective types " ^
				 (string_of_type_expression e1_type) ^
				 " and " ^
				 (string_of_type_expression e2_type))))
		      
		|  ftype -> raise
		     (Exc_typeexpression
			(op ^
			   " is used as an infix operator but has " ^
			   "non-functional type" ^
			   (string_of_type_expression ftype)))
	    with
		Not_found ->
		  raise (Exc_typeexpression
			   ("infix operator " ^ op ^ " is not declared"))




		    
		    

	    
	

