(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

(* exception raised in case of a parsing or analysis error *)
exception Exc_prouve of string


(************************************************************)
(*           Abstract syntax of PROUVE files                *)
(************************************************************)

module AbstractSyntax =
struct
  (* type expressions as defined in module Types *)
  type type_expression
    = Types.type_expression =
      Typeconst of string
      | Typevar of string
      | Typeapply of string * type_expression list
      | Typearrow of type_expression * type_expression
      | Typetuple of type_expression list
      | Typeundef
  let string_of_type_expression = Types.string_of_type_expression 
  let is_simple_type = Types.is_simple_type
  let smaller_type = Types.smaller_type
    
  (* signature declarations as defined in module Signatures  *)
  type simple_signature = Signatures.simple_signature
  type signature = Signatures.signature
  let print_signature = Signatures.print_signature
  let default_signature = Defaultsignature.default_signature
  let list_of_constants = Signatures.list_of_constants

  (* kinds of declarations as defined in module Declarations *)
  type declarationkind = Declarations.declarationkind =
      Logical | Mutable | ProcessKind
  let string_of_declarationkind = Declarations.string_of_declarationkind

  (* variable declaration as defined in module Declarations *)
  type declaration = Declarations.declaration


  (* expressions as defined in module Expressions *)

  type var_kind = Expressions.var_kind =
    | Var_unknown    (* kind is not yet known *)
    | Var_value      (* can only be an l value *)
    | Var_logical    (* expression which is a logical variable *)
    | Var_mutable    (* expression which is a mutable variable *)
    | Var_of_process (* access to variable of a process *)
    | Var_pid        (* variable denoting a process *)

  type info_exp = Expressions.info_exp = {
    exp_type : Types.type_expression;
    exp_varkind : var_kind;
    exp_hasvalue : bool
  }

  type expression = Expressions.expression =
    Ident of info_exp * string
  | Int of info_exp * int
  | Apply of info_exp * string * expression list
  | Infix of info_exp * string * expression * expression
  | Inverse of info_exp * expression
  | Tuple of info_exp * expression list
  | Liste of info_exp * expression list
  | AssocList of info_exp * (expression*expression) list

  let string_of_expression = Expressions.string_of_expression
  let string_of_expression_list = Expressions.string_of_expression_list
  let is_ground = Expressions.is_ground
  let get_info_exp = Expressions.get_info_exp
  let vars = Expressions.vars

  (* axioms as defined in module Axioms *)
  type axioms = Axioms.axioms
  let is_a_axiom = Axioms.is_a_axiom
  let is_c_axiom = Axioms.is_c_axiom
      
  (* patterns as defined in moule Expressions *)
  type pattern = Expressions.pattern
  let string_of_pattern = Expressions.string_of_pattern

  (* instructions as defined in module Roles *)
  type instruction =
      Roles.instruction =
    IBlock of string*block
  | ISend of string*expression
  | IRecv of string*pattern
  | IFail of string
  | IAssign of string * string * expression
  | INew of string * string
  | IMakeKey of string  * expression * string
  | IMatch of string * expression * pattern
  | IConditional of string * expression * instruction list * instruction list
  | IChoice of string * instruction list list
  | ICase of string *
      expression * (pattern * instruction list) list * instruction list
  and block = declaration list * instruction list

  (* a roles as defined in module Roles *)
  type role = Roles.role = Role of string * declaration list * block
  let print_role = Roles.print_role
  let get_form_parlist = Roles.get_form_parlist

  (* variable declarations with explicit skolemisation, as defined in *)
  (* module Declarations.                                             *)
  type sk_declaration = Declarations.sk_declaration

  (* extended integers as defined in module Intervals *)
  type omint = Intervals.omint = OmInt of int | Omega

  (* integer intervals as defined in module Intevals *)
  type interval = Intervals.interval

  (* scenarios as defined in module Scenarios *)
  type scenario
    = Scenarios.scenario =
      SCall of string * string * expression list
      | SNew of string * expression
      | SForever of string * scenario list
      | SParallel of string * scenario list
      | SSequence of string * scenario list
      | SAssign of string * expression * expression
      | SConditional of string * expression * scenario list * scenario list
      | SForall of string * declaration list * interval list * scenario
      | SExists of string * declaration list * scenario
      | SBlock of string * declaration list * scenario list
  let print_scenario = Scenarios.print_scenario

  (* specifications as defined in module Specs *)
  type spec = Specs.spec
  let print_spec = Specs.print_spec
  let parse_spec filename =
    try Specparser.parse_spec filename
    with Specparser.Exc_specparser s -> raise (Exc_prouve s)
end
  
(*****************************************************************)
(* Module for Logical Formulas and Assertions                    *)
(*****************************************************************)

module Logic = struct
  type assertion =
      string * string list * Formulas.formula * Formulas.formula

  let print_assertion = Assertions.print_assertion
	
  type formula = Formulas.formula =
      Ftrue
    | Fatom of Expressions.expression
    | Fsecret of Expressions.expression
    | Fprocessat of string * string
    | Fscenarioat of string
    | Fand of formula * formula
    | For of formula * formula
    | Fimplies of formula * formula
    | Fnot of formula
    | Fforall of Declarations.declaration list * formula
    | Fexists of Declarations.declaration list * formula
    | Funique of Declarations.declaration list * formula
    | Fbox of formula

  let string_of_formula=Formulas.string_of_formula

  let parse_assertion filename =
    try Assparser.parse_assertion filename
    with Assparser.Exc_assparse s -> raise (Exc_prouve s)

end

(************************************************************************)
(* Term Rewriting Systems                                               *)
(************************************************************************)

module TermRewriting = struct
  type trs = ( Expressions.expression * Expressions.expression ) list
  let trs_to_string = Trs.trs_to_string
  type substitution = (string * Expressions.expression) list
  let apply_subst = Trs.apply_subst
  let string_of_substitution = Trs.string_of_substitution
end

(************************************************************************)
(* Miscellaneous                                                        *)
(************************************************************************)

module Misc = struct
  let remove_double_occurrences = Utils.singlyfy
end
