(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Intervals
open Expressions
open Declarations


(********* abstract syntax *******************************************)

(* raw abstract syntax *)
type rawscenario =
    SRawLabeled of string * rawscenario
  | SRawCall of string * expression list
  | SRawNew of expression
  | SRawForever of rawscenario list
  | SRawParallel of rawscenario list
  | SRawSequence of rawscenario list
  | SRawAssign of expression * expression
  | SRawConditional of expression * rawscenario list * rawscenario list
  | SRawForall of declaration list * interval list * rawscenario
  | SRawExists of declaration list * rawscenario
  | SRawBlock of declaration list * rawscenario list

(* processed abstract syntax. The first argument is the label of a   *)
(* sub-scenario.                                                     *)
type scenario =
    SCall of string * string * expression list
  | SNew of string * expression
  | SForever of string * scenario list
  | SParallel of string * scenario list
  | SSequence of string * scenario list
  | SAssign of string * expression * expression
  | SConditional of string * expression * scenario list * scenario list
  | SForall of string * declaration list * interval list * scenario
  | SExists of string * declaration list * scenario
  | SBlock of string * declaration list * scenario list


(*************** processing *******************************************)

let rec process_scenario label = function
    SRawLabeled(newlabel,scene) -> process_scenario newlabel scene
  | SRawCall(procname,args) -> SCall(label,procname,args)
  | SRawNew(exp) -> SNew(label,exp)
  | SRawForever(instlist) -> SForever(label, process_list label instlist)
  | SRawParallel(instlist) -> SParallel(label, process_list label instlist)
  | SRawSequence(instlist) -> SSequence(label, process_list label instlist)
  | SRawAssign(lexp,rexp) -> SAssign(label,lexp,rexp)
  | SRawConditional(cond,thenpart,elsepart) ->
      SConditional(label,cond,
		   (process_list (label^"_thn") thenpart),
		   (process_list (label^"_els") elsepart))
  | SRawExists(decllist,scene) ->
      SExists(label, decllist, process_scenario (label^"_bdy") scene)
  | SRawForall(decllist,intervals,scene) ->
      SForall(label, decllist, intervals,
	      process_scenario (label^"_bdy") scene)
  | SRawBlock(decllist, instlist) -> 
      SBlock(label, decllist, process_list label instlist)
and process_list label =
  Utils.mapindex (function n ->  process_scenario(label^"_"^(string_of_int n)))

(*************** printing *********************************************)

(* (print_sc showtype oc prefix inst) prints to the output channel oc the *)
(* instruction inst, possibly on several lines, terminated by a "\n".     *)
(* "prefix" is the indention of lines.                                    *)
(* "showtype" indicates whether the type of expressions is to be printed  *)
let rec print_sc showtype oc prefix = function
    SCall(label,rulename, args) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc rulename;
      output_char oc '(';
      output_string oc (string_of_expression_list showtype args);
      output_string oc ")"
  | SNew(label,exp) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "new";
      output_char oc '(';
      output_string oc (string_of_expression showtype exp);
      output_string oc ")"
  | SAssign(label,lexp,rexp) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc (string_of_expression showtype lexp);
      output_string oc " := ";
      output_string oc (string_of_expression showtype rexp);
 | SConditional(label,cond,thenpart,elsepart) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "if ";
      output_string oc (string_of_expression showtype cond);
      output_string oc " then\n";
      print_sc_list showtype oc (prefix^"  ") thenpart;
      output_string oc prefix;
      output_string oc "else\n";
      print_sc_list showtype oc (prefix^"  ") elsepart;
      output_string oc prefix;
      output_string oc "fi"
 | SForever (label,sl) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "forever\n";
      print_sc_list showtype oc (prefix^"  ") sl;
      output_string oc prefix;
      output_string oc "end";
 | SSequence (label,sl) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "begin\n";
      print_sc_list showtype oc (prefix^"  ") sl;
      output_string oc prefix;
      output_string oc "end";
 | SParallel (label,sl) ->
      output_string oc prefix;
      output_string oc (label^": ");
      output_string oc "parallel\n";
      print_sc_par_list showtype oc (prefix^"  ") sl;
      output_string oc "end";
 | SExists (label,dl,inst) ->
     output_string oc prefix;
     output_string oc (label^": ");
     output_string oc "exists ";
     output_string oc (string_of_parlist dl);
     output_string oc " .\n";
     print_sc showtype oc (prefix^"  ") inst;
 | SForall (label,dl,intervals,inst) ->
     output_string oc prefix;
     output_string oc (label^": ");
     output_string oc "forall ";
     output_string oc (string_of_parlist dl);
     output_string oc (string_of_interval_list intervals);
     output_string oc " .\n";
     print_sc showtype oc (prefix^"  ") inst;
 | SBlock (label, dl, instlist) ->
     output_string oc prefix;
     output_string oc (label^": ");
     output_string oc "declare\n";
     List.iter
       (function decl ->
	  output_string oc (prefix^"  ");
	  output_string oc (string_of_declaration decl);
	  output_string oc ";\n")
       dl;
     output_string oc prefix;
     output_string oc "begin\n";
     print_sc_list showtype oc (prefix^"  ") instlist;
     output_string oc prefix;
     output_string oc "end";

(* (print_sc_list outchannel prefix sclist) outputs to                *)
(* the output channel oc the list of scene instructions sclist, where *)
(* every instruction is prefixed by prefix and terminated by ";\n"     *)
and print_sc_list showtype oc prefix =
  List.iter
    (function inst ->
       print_sc showtype oc prefix inst;
       output_string oc ";\n")
    
(* (print_sc_list outchannel prefix sclist) outputs to                 *)
(* the output channel oc the list of scene instructions sclist, where  *)
(* every instruction is prefixed by prefix^"  " and terminated by "\n" *)
(* except for the first instruction which is prefixed by prefix^"| "   *)
and print_sc_par_list showtype oc prefix = function
    [] -> ()
  | h::r ->
      print_sc showtype oc (prefix^"  ") h;
      output_string oc "\n";
      List.iter
	(function inst ->
	   output_string oc (prefix^"|\n");
	   print_sc showtype oc (prefix^"  ") inst;
	   output_string oc "\n")
	r
    

let print_scenario showtype oc = print_sc showtype oc ""

(*************** extract integer constants ****************************)
let rec ic = function 
    SCall(_,_,el) -> List.flatten (List.map Expressions.int_constants el)
  | SForever (_,l) | SSequence (_,l) | SParallel (_,l) | SBlock(_,_,l) ->
      ic_list l
  | SAssign(_,le,re) ->
      (Expressions.int_constants le) @ (Expressions.int_constants re)
  | SNew(_) -> []
  | SConditional(_,cond,thenl,elsel) ->
      (Expressions.int_constants cond) @
      (ic_list thenl) @ (ic_list elsel)
  | SForall(_,_,_,s) | SExists(_,_,s) -> ic s
and ic_list l = List.flatten (List.map ic l)

let int_constants l = Utils.singlyfy (ic l)


(*************** collect checkpoints *********************************)

let rec checkpoints = function
    SCall(l,_,_) | SNew(l,_) | SAssign(l,_,_) -> [l]
  | SForever(l,slist) | SParallel(l,slist) | SSequence(l,slist) 
  | SBlock(l,_,slist) ->
      l::(checkpoints_list slist)
  | SExists(l,_,scene) | SForall(l,_,_,scene) ->
      l::(checkpoints scene)
  | SConditional(l,_,thenpart,elsepart) ->
      l::(checkpoints_list thenpart)@(checkpoints_list elsepart)
and checkpoints_list slist = 
  List.fold_right (fun x y -> (checkpoints x)@y) slist []

let gather_checkpoints scenario = "start"::"end"::(checkpoints scenario)



exception Exc_scenario of string

