(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

open Declarations
open Roles

(* (variant x l) returns a "variant" of the string x not occurring in l *)
let variant x l =
  let rec variant_aux x n l =
    let newx = x^"_"^(string_of_int n)
    in
      if List.mem newx l
      then variant_aux x (n+1) l
      else newx
  in
    variant_aux x 0 l
  
(* (unnest_inst decl inst), where decl is a declaration list and inst  *)
(* an instruction such that all free variables of inst are declated in *)
(* decl, yields a pair consisting of a                                 *)
(* declaration decl' and an instruction inst' such that:               *)
(* 1.) decl and decl' are disjoint                                     *)
(* 2.) inst' does not contain variable declarations                    *)
(* 3.) the block                                                       *)
(*         declare dl begin inst end                                   *)
(*     is alpha-equivalent to                                          *)
(*         declare dl; dl' begin inst' end                             *)
(* unnest_instlist does the same for instruction lists.                *)
(* unnest_caselist does the same for case lists.                       *)

let rec unnest_inst decl = function
    IBlock(label,(dl,il)) ->
      let (dl',il') = unnest_ilist (decl@dl) il
	(* decl and dl may be non-disjoint. Hence, we have to rename all *)
	(* variables in dl and il' which also occur in decl. We replace  *)
	(* these variables by new ones not occurring in decl or dl'.     *)
      in let domain_exclude =
	  (* var's to be renamed when defined by dl *)
	  (* by recursion hypothesis, the domains of dl' and of dl are   *)
	  (* already disjoint.                                           *)
	  (List.map fst decl)
      in let codomain_exclude =
	  (* varnames to avoid in the renaming: variables defined in  *)
	  (* decl or in dl'.                                          *)
	  domain_exclude @ (List.map fst dl')
      in let alpha =
	  (* alpha is the renaming *)
	  List.fold_right
	    (fun (x,_) l -> 
	       if List.mem x domain_exclude
	       then (x, (variant x codomain_exclude)):: l
	       else l)
	    dl
	    []
      in ((rename_dlist alpha dl)@dl',
	  IBlock(label,([], rename_ilist alpha il')))
  | IConditional(label,cond,thenpart,elsepart) ->
      let (dlthen,ilthen) = unnest_ilist decl thenpart
      in let (dlelse,ilelse) = unnest_ilist (decl@dlthen) elsepart
      in (dlthen@dlelse, IConditional(label,cond,ilthen,ilelse))
  | IChoice(label,choicelist) ->
      let (dl',choicelist') = unnest_ilistlist decl choicelist
      in (dl',IChoice(label,choicelist'))
  | ICase(label,expr,caselist,elsepart) ->
      let (dl',caselist') = unnest_caselist decl caselist
      in let (dl'',elsepart') = unnest_ilist (decl@dl') elsepart
      in ((dl'@dl''), ICase(label,expr,caselist',elsepart'))
  | (ISend _|IRecv _|IFail _|IAssign _|INew _|IMatch _|IMakeKey _) as inst ->
      [],inst

and unnest_ilist decl = function
    [] -> [],[]
  | h::r -> 
      let (dlh,ih)= unnest_inst decl h
      in let (dlr,ilr) = unnest_ilist (decl@dlh) r
      in (dlh@dlr, ih::ilr) 

and unnest_ilistlist decl = function
    [] -> [],[]
  | h::r -> 
      let (dlh,ih)= unnest_ilist decl h
      in let (dlr,ilr) = unnest_ilistlist (decl@dlh) r
      in (dlh@dlr, ih::ilr) 

and unnest_caselist decl = function
    [] -> [], []
  | (pat,h)::r ->  let (dlh,ilh)= unnest_ilist decl h
      in let (dlr,ilr) = unnest_caselist (decl@dlh) r
      in (dlh@dlr, (pat,ilh)::ilr) 

let unnest (Role (rolename,decllist,(dl,il))) = 
  let (dl',il') = unnest_ilist (decllist@dl) il
  in Role(rolename, decllist, (dl@dl', il'))


let simplify (signature, axioms, rolelist, vars, scene) =
  (signature, axioms, List.map unnest rolelist, vars, scene)
