(**************************************************************************)
(*  ProuveParser: parser frontend for the PROUVÉ language                *)
(*                                                                        *)
(*  Copyright (C) 2004-2009 Ralf Treinen                                  *)
(*                                                                        *)
(*  This library is free software; you can redistribute it and/or         *)
(*  modify it under the terms of the GNU Lesser General Public            *)
(*  License as published by the Free Software Foundation; either          *)
(*  version 2.1 of the License, or (at your option) any later version.    *)
(*                                                                        *)
(*  This library is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *)
(*  Lesser General Public License for more details.                       *)
(*                                                                        *)
(*  You should have received a copy of the GNU Lesser General Public      *)
(*  License along with this library; if not, write to the Free Software   *)
(*  Foundation, Inc., 59 Temple Place, Suite 330,                         *)
(*  Boston, MA  02111-1307  USA                                           *)
(*                                                                        *)
(*  Ralf Treinen <treinen@pps.jussieu.fr>                                 *)
(*  Laboratoire Preuves, Programmes et Systèmes (PPS),                   *)
(*  Université Paris Diderot, Paris, France                              *)
(*  http://www.pps.jussieu.fr/~treinen                                    *)
(*                                                                        *)
(*  This software was mainly written while the author was with the        *)
(*  Laboratoire Spécification et Vérification (LSV),                    *)
(*  ENS de Cachan, Cachan, France                                         *)
(**************************************************************************)

(***************************************************************************)
(* Types and polymorphic type expressions                                  *)
(***************************************************************************)

(* Type of all type expression, including the internal type Undef      *)
(* The constructor Typeundef is an anonymous type variable that may    *)
(* be used as the argument of a type constructor, like list (to denote *)
(* the type of empty list).                                            *)

type type_expression =
    Typeconst of string
  | Typevar   of string
  | Typeapply of string * type_expression list
  | Typearrow of type_expression * type_expression
  | Typetuple of type_expression list
  | Typeundef

(* test whether a type is simple, that is only built of type constants, *)
(* tupling, and arrows.                                                 *)
val is_simple_type: type_expression -> bool

(* convert a type expression, or a list of type expressions,        *)
(* into a string                                                    *)
val string_of_type_expression: type_expression -> string
val string_of_type_expression_list: type_expression list -> string

(* check that only predefined type constants, and predefined type    *)
(* constructors are used with the right arity                        *)
exception Exc_typecons of string
val check_type_constructors: type_expression -> unit

(* test whether a type expression is ground, that is whether it does    *)
(* not contain a named type variable. However, a ground type expression *)
(* may contain the anonymous type variable Typeundef.                   *)
val is_ground: type_expression -> bool

(* check that a type_expression is well-formed:                     *)
(* - it passes test check_type_constructors                         *)
(* - it passes test is_ground                                       *)
(* - either arrow-free, or of the form (t1 -> t2) where t1 and t2   *)
(*   are arrow-free.                                                *)     
exception Wellformed_type of string                            
val well_formed: type_expression -> unit

(* check that a type expression does not contain an arrow *)
val arrow_free: type_expression -> bool

(* the set of predefined type constants *)
val predefined_type_constants: string list



(*********** Substitution and matching ***************************************)

(* (type_match pattern type), where pattern is an arrow-free type expression *)
(* and type is an arrow-free and variable-free type expression returns       *)
(* a matching substitution sub such that sub(pattern) = type if such a       *)
(* substitution exists, otherwise raises the exception Nomatch_type          *)
exception Nomatch_type 
val type_match:
  type_expression -> type_expression ->  (string*type_expression) list

(* (apply_type_subst subst te) applies the substitution subst to the *)
(* type expression te.                                               *)
val apply_type_subst:
  (string*type_expression) list -> type_expression -> type_expression


(********* Subtyping ***************************************************)

(* (sup_type t1 t2) returns the supremum of two type expressions      *)
(* in the flat cpo with bottom element Typeundef.                     *)
val sup_type: type_expression -> type_expression -> type_expression

(* (has_inf_type t1 t2) tests whether the two type expressions t1 and *)
(* t2, which are arrow-free and ground, have a common infimum.        *)
val has_inf_type: type_expression -> type_expression -> bool

(* (smaller_type t1 t2), where t1 and t2 are ground and arrow_free, tests *)
(* whether t1 is smaller than t2.                                         *)
val smaller_type: type_expression -> type_expression -> bool


